/*-----------------------------------------------------------------------------
 * $RCSfile: Socket.c,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if (!defined(_WIN32)) || defined(_DEBUG)
static  const   char    CVS_ID[]        =       "$Id: Socket.c,v 1.6 2001/12/18 02:00:49 bandou Exp $";
#endif  /* _DEBUG */

#include "Socket.h"

/*
=begin
= OpenSOAP Socket class
=end
 */

#if defined(NONGCC_WIN32)

#define SOCKET_INITIALIZE(s) ((s) = INVALID_SOCKET)
#define IS_VALID_SOCKET(s)  ((s) != INVALID_SOCKET)

static
int
CLOSE_SOCKET(SOCKET_TYPE s) {
    int ret = OPENSOAP_NO_ERROR;

    if (closesocket(s) == SOCKET_ERROR) {
        int c_err = WSAGetLastError();
        if (c_err == WSANOTINITIALISED) {
            ret = OPENSOAP_IO_ERROR;
        }
        else if (c_err == WSAENETDOWN) { /* network subsystem has
                                          * failed */
            ret = OPENSOAP_IO_ERROR;
        }
        else if (c_err == WSAENOTSOCK) { /* descriptor is not a socket */
            ret = OPENSOAP_IO_ERROR;
        }
        else if (c_err == WSAEINPROGRESS) { /* in progress */
            ret = OPENSOAP_IO_ERROR;
        }
        else if (c_err == WSAEINTR) { /* close interrupted */
            ret = OPENSOAP_IO_ERROR;
        }
        else if (c_err == WSAEWOULDBLOCK) { /* */
            ret = OPENSOAP_IO_ERROR;
        }
        else {
            ret = OPENSOAP_IO_ERROR;
        }
    }

    return ret;
}

/* Winsock initialize */
static
const WORD
winsockRequestVersion = MAKEWORD(1,1);

typedef struct {
    OpenSOAPObject super;
    int errorWSAStartup;
    WSADATA wsaData;
} OpenSOAPSocketSystem;

typedef OpenSOAPSocketSystem *OpenSOAPSocketSystemPtr;

static
OpenSOAPSocketSystemPtr
theSocketSystem = NULL;

static
int
OpenSOAPSocketSystemFree(OpenSOAPObjectPtr obj) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (obj) {
        OpenSOAPSocketSystemPtr s_s = (OpenSOAPSocketSystemPtr)obj;
        if (!s_s->errorWSAStartup) {
            WSACleanup();
        }
        ret = OpenSOAPObjectReleaseMembers(obj);
        if (OPENSOAP_SUCCEEDED(ret)) {
            free(obj);
            theSocketSystem = NULL;
        }
    }

    return ret;
}

static
int
OpenSOAPSocketSystemCreate(OpenSOAPSocketSystemPtr * /* [out] */ s_s) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (s_s) {
        ret = OPENSOAP_MEM_BADALLOC;
        *s_s = malloc(sizeof(OpenSOAPSocketSystem));
        if (*s_s) {
            ret = OpenSOAPObjectInitialize((OpenSOAPObjectPtr)*s_s,
                                           OpenSOAPSocketSystemFree);
            if (OPENSOAP_SUCCEEDED(ret)) {
                /* Initialize WinSock.dll */
                (*s_s)->errorWSAStartup
                    = WSAStartup(winsockRequestVersion,
                                 &(*s_s)->wsaData);
                if ((*s_s)->errorWSAStartup) {
                    /* Winsock Initialize error */
                    int wsa_error
                        = (*s_s)->errorWSAStartup;
                    (*s_s)->errorWSAStartup = 0;
                    switch (wsa_error) {
                      case WSASYSNOTREADY:
                          ret = OPENSOAP_IO_ERROR;
                          break;
                      case WSAVERNOTSUPPORTED:
                          ret = OPENSOAP_IO_ERROR;
                          break;
                      case WSAEINPROGRESS:
                          ret = OPENSOAP_IO_ERROR;
                          break;
                      case WSAEPROCLIM:
                          ret = OPENSOAP_IO_ERROR;
                          break;
#if defined(WSAFAULT)
                      case WSAFAULT:
                          ret = OPENSOAP_IO_ERROR;
                          break;
#endif /* WSAFAULT */
                      default:
                          ret = OPENSOAP_IO_ERROR;
                          break;
                    }
                }
                else if ((*s_s)->wsaData.wVersion != winsockRequestVersion) {
                    /* winsock version mismatch */
                    ret = OPENSOAP_IO_ERROR;
                }
                if (OPENSOAP_FAILED(ret)) {
                    OpenSOAPSocketSystemFree((OpenSOAPObjectPtr)*s_s);
                    *s_s = NULL;
                }
            }
            else {
                free(*s_s);
                *s_s = NULL;
            }
        }
    }

    return ret;
}

static
int
OpenSOAPSocketSystemInitialize() {
    int ret
        = (!theSocketSystem)
        ? OpenSOAPSocketSystemCreate(&theSocketSystem)
        :  OpenSOAPObjectRetain((OpenSOAPObjectPtr)theSocketSystem);

    return ret;
}

static
void
OpenSOAPSocketSystemUltimate() {
    if (theSocketSystem) {
        OpenSOAPObjectRelease((OpenSOAPObjectPtr)theSocketSystem);
    }
}

#else
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define SOCKET_INITIALIZE(s) ((s) = -1)
#define IS_VALID_SOCKET(s)  ((s) != -1)

static
int
CLOSE_SOCKET(SOCKET_TYPE s) {
    int ret = OPENSOAP_NO_ERROR;

    if (!close(s)) {
        int c_err = errno;
        if (c_err == EBADF) { /* invalid fd */
            ret = OPENSOAP_IO_ERROR;
        }
        else if (c_err == EINTR) { /* close interrupt */
            ret = OPENSOAP_IO_ERROR;
        }
        else if (c_err == EIO) { /* I/O error */
            ret = OPENSOAP_IO_ERROR;
        }
        else {
            ret = OPENSOAP_IO_ERROR;
        }
    }

    return ret;
}


static
int
OpenSOAPSocketSystemInitialize() {
    return OPENSOAP_NO_ERROR;
}

static
void
OpenSOAPSocketSystemUltimate() {
}

#endif

/*
=begin
--- function#OpenSOAPSocketRelease(s)
    OpenSOAP Socket JB

    :Parameters
      :OpenSOAPSocketPtr  [in] ((|s|))
        OpenSOAP Socket
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketReleaseMembers(OpenSOAPSocketPtr s) {
    int ret = OpenSOAPSocketClose(s);

    if (OPENSOAP_SUCCEEDED(ret)) {
        OpenSOAPSocketSystemUltimate();
        ret = OpenSOAPStreamReleaseMembers((OpenSOAPStreamPtr)s);
    }

    return ret;
}

static
int
OpenSOAPSocketFree(OpenSOAPObjectPtr obj) {
    OpenSOAPSocketPtr s = (OpenSOAPSocketPtr)obj;
    int ret = OpenSOAPSocketReleaseMembers(s);

    if (OPENSOAP_SUCCEEDED(ret)) {
        free(obj);
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPSocketReadImpl(st, buf, buf_sz)
    OpenSOAP Socket ǂݍށB

    :Parameters
      :OpenSOAPStreamPtr  [in] ((|st|))
        OpenSOAP Socket
      :unsigned char * [out] buf,
        ǂݍ݃obt@̐擪|C^B
      :size_t * [in, out] buf_sz
        ǂݍ݃obt@̃TCYBďo͓ǂݍ񂾃TCYԂB
    :Returns
      :int
      
=end
 */
static
int
OpenSOAPSocketReadImpl(OpenSOAPStreamPtr /* [in] */ st,
                       unsigned char * /* [out] */ buf,
                       size_t * /* [in, out] */ buf_sz) {
    OpenSOAPSocketPtr s = (OpenSOAPSocketPtr)st;
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (s && buf && buf_sz && *buf_sz) {
        ret = OPENSOAP_IO_ERROR; /* not open socket */
        if (IS_VALID_SOCKET(s->socket)) {
            unsigned char *buf_end = buf + *buf_sz;
            unsigned char *buf_i = buf;

            ret = OPENSOAP_NO_ERROR;
            while (buf_i < buf_end) {
                int recv_sz
                    = recv(s->socket,
                           buf_i,
                           buf_end - buf_i,
                           0);
                if (recv_sz < 0) {
                    ret = OPENSOAP_IO_ERROR;
                    break;
                }
                else if (recv_sz == 0) {
                    break;
                }
                buf_i += recv_sz;
            }
            if (OPENSOAP_SUCCEEDED(ret)) {
                *buf_sz = buf_i - buf;
            }
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPSocketWriteImpl(st, buf, buf_sz)
    OpenSOAP Socket ֏oB

    :Parameters
      :OpenSOAPStreamPtr  [in] ((|st|))
        OpenSOAP Socket
      :const unsigned char * [in] buf,
        ݃f[^̐擪|C^B
      :size_t * [in, out] buf_sz
        ݃f[^̃TCYBďo͏񂾃TCYԂB
    :Returns
      :int
      
=end
 */
static
int
OpenSOAPSocketWriteImpl(OpenSOAPStreamPtr /* [in] */ st,
                    const unsigned char * /* [in] */ buf,
                    size_t * /* [in, out] */ buf_sz) {
    OpenSOAPSocketPtr s = (OpenSOAPSocketPtr)st;
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (s && buf && buf_sz && *buf_sz) {
        ret = OPENSOAP_IO_ERROR; /* not open socket */
        if (IS_VALID_SOCKET(s->socket)) {
            const unsigned char *buf_end = buf + *buf_sz;
            const unsigned char *buf_i = buf;
            ret = OPENSOAP_NO_ERROR;
            while (buf_i < buf_end) {
                int send_sz = send(s->socket,
                                   buf_i,
                                   buf_end - buf_i,
                                   0);
                if (send_sz < 0) {
                    ret = OPENSOAP_IO_ERROR;
                    break;
                }
                else if (send_sz == 0) {
                    break;
                }
                buf_i += send_sz;
            }
            if (OPENSOAP_SUCCEEDED(ret)) {
                *buf_sz = buf_i - buf;
            }
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPSocketInitialize(s, free_func, read_func, write_func)
    OpenSOAP Socket member initialize.

    :Parameters
      :OpenSOAPSocketPtr * [out] ((|s|))
        OpenSOAP Socket pointer
      :OpenSOAPObjectFreeFunc [in] ((|free_func|))
        OpenSOAP Object free function
      :OpenSOAPStreamReadFunc [in] ((|read_func|))
        OpenSOAP Stream read function.
        If NULL, use default read function.
      :OpenSOAPStreamWriteFunc [in] ((|write_func|))
        OpenSOAPStream write function.
        If NULL, use default write function.
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketInitialize(OpenSOAPSocketPtr /* [in, out] */ s,
                         OpenSOAPObjectFreeFunc /* [in] */ free_func,
                         OpenSOAPStreamReadFunc  /* [in] */ read_func,
                         OpenSOAPStreamWriteFunc /* [in] */ write_func) {
    int ret = OpenSOAPSocketSystemInitialize();
    
    if (OPENSOAP_SUCCEEDED(ret)) {
        OpenSOAPStreamPtr st = (OpenSOAPStreamPtr)s;
        if (!read_func) {
            read_func = OpenSOAPSocketReadImpl;
        }
        if (!write_func) {
            write_func = OpenSOAPSocketWriteImpl;
        }
        ret = OpenSOAPStreamInitialize(st,
                                       free_func,
                                       read_func,
                                       write_func);
        if (OPENSOAP_SUCCEEDED(ret)) {
            SOCKET_INITIALIZE(s->socket);
        }
    }
    
    return ret;
}

/*
=begin
--- function#OpenSOAPSocketCreate(s)
    OpenSOAP Socket instance create

    :Parameters
      :OpenSOAPSocketPtr * [out] ((|s|))
        OpenSOAP Socket pointer
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketCreate(OpenSOAPSocketPtr * /* [out] */ s) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (s) {
        *s = malloc(sizeof(OpenSOAPSocket));
        if (*s) {
            ret = OpenSOAPSocketInitialize(*s,
                                           OpenSOAPSocketFree,
                                           NULL, NULL);
            if (OPENSOAP_FAILED(ret)) {
                free(*s);
                *s = NULL;
            }
        }
        else {
            ret = OPENSOAP_MEM_BADALLOC;
        }
    }

    return ret;
}

      
/*
=begin
--- function#OpenSOAPSocketRelease(s)
    OpenSOAP Socket JB

    :Parameters
      :OpenSOAPSocketPtr  [in] ((|s|))
        OpenSOAP Socket
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketRelease(OpenSOAPSocketPtr /* [in] */ s) {
    return OpenSOAPObjectRelease((OpenSOAPObjectPtr)s);
}

/*
=begin
--- function#OpenSOAPSocketOpen(s, node, service)
    OpenSOAP Socket I[vB

    :Parameters
      :OpenSOAPSocketPtr  [in] ((|s|))
        OpenSOAP Socket
      :int [in] domain
        ʐMsȂhC(domain)B
      :int [in] type
        ʐM(semantics)B
      :int [in] protocol
        \PbgɂĎgpŗL̃vgRB
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketOpen(OpenSOAPSocketPtr /* [in, out] */ s,
                   int /* [in] */ domain,
                   int /* [in] */ type,
                   int /* [in] */ protocol) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (s) {
        SOCKET_TYPE soc = socket(domain, type, protocol);
        ret = OPENSOAP_IO_ERROR;
        if (IS_VALID_SOCKET(soc)) {
            s->socket = soc;
            ret = OPENSOAP_NO_ERROR;
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPSocketOpen(s, node, service)
    OpenSOAP Socket I[vB

    :Parameters
      :OpenSOAPSocketPtr  [in] ((|s|))
        OpenSOAP Socket
      :int [in] domain
        ʐMsȂhC(domain)B
      :int [in] type
        ʐM(semantics)B
      :int [in] protocol
        \PbgɂĎgpŗL̃vgRB
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketConnect(OpenSOAPSocketPtr /* [in, out] */ s,
                      const struct sockaddr * /* [in] */ serv_addr,
                      size_t /* [in] */ addrlen) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (s) {
        ret = OPENSOAP_IO_ERROR; /* socket not open */
        if (IS_VALID_SOCKET(s->socket)) {
            if (!connect(s->socket, serv_addr, addrlen)) {
                ret = OPENSOAP_NO_ERROR;
            }
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPSocketClose(s)
    OpenSOAP Socket N[YBڑؒfB

    :Parameters
      :OpenSOAPSocketPtr  [in] ((|s|))
        OpenSOAP Socket
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketClose(OpenSOAPSocketPtr /* [in] */ s) {
    int ret = OPENSOAP_PARAMETER_BADVALUE;

    if (s) {
        ret = OPENSOAP_NO_ERROR;
        if (IS_VALID_SOCKET(s->socket)) {
            ret = CLOSE_SOCKET(s->socket);
            SOCKET_INITIALIZE(s->socket);
        }
    }

    return ret;
}

/*
=begin
--- function#OpenSOAPSocketRead(s, buf, buf_sz)
    OpenSOAP Socket ǂݍށB

    :Parameters
      :OpenSOAPSocketPtr  [in] ((|s|))
        OpenSOAP Socket
      :unsigned char * [out] buf,
        ǂݍ݃obt@̐擪|C^B
      :size_t * [in, out] buf_sz
        ǂݍ݃obt@̃TCYBďo͓ǂݍ񂾃TCYԂB
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketRead(OpenSOAPSocketPtr /* [in] */ s,
                   unsigned char * /* [out] */ buf,
                   size_t * /* [in, out] */ buf_sz) {
    int ret = OpenSOAPStreamRead((OpenSOAPStreamPtr)s,
                                 buf,
                                 buf_sz);

    return ret;
}

/*
=begin
--- function#OpenSOAPSocketWrite(s, buf, buf_sz)
    OpenSOAP Socket ֏oB

    :Parameters
      :OpenSOAPSocketPtr  [in] ((|s|))
        OpenSOAP Socket
      :const unsigned char * [in] buf,
        ݃f[^̐擪|C^B
      :size_t * [in, out] buf_sz
        ݃f[^̃TCYBďo͏񂾃TCYԂB
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketWrite(OpenSOAPSocketPtr /* [in] */ s,
                    const unsigned char * /* [in] */ buf,
                    size_t * /* [in, out] */ buf_sz) {
    int ret = OpenSOAPStreamWrite((OpenSOAPStreamPtr)s,
                                  buf,
                                  buf_sz);

    return ret;
}

/*
=begin
--- function#OpenSOAPSocketReadToDelm(s, delm, delm_sz, buf)
    OpenSOAP Socket f~^܂œǂݍށB

    :Parameters
      :OpenSOAPSocketPtr  [in] ((|s|))
        OpenSOAP Socket
      :const unsigned char * [in] delm
        f~^f[^B
      :size_t [in] delm_sz
        f~^f[^񒷂B
      :OpenSOAPByteArrayPtr [out] buf
        ǂݍ݃obt@B
    :Returns
      :int
      
=end
 */
extern
int
OPENSOAP_API
OpenSOAPSocketReadToDelm(OpenSOAPSocketPtr /* [in] */ s,
                         const unsigned char * /* [in] */ delm,
                         size_t /* [in] */ delm_sz,
                         OpenSOAPByteArrayPtr /* [out] */ buf) {
    int ret = OpenSOAPStreamReadToDelm((OpenSOAPStreamPtr)s,
                                       delm,
                                       delm_sz,
                                       buf);

    return ret;
}

